from django.urls import resolve

from django_glue.conf import settings
from django_glue.glue import clean_glue_session


class GlueMiddleware(object):
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        current_url = resolve(request.path_info).url_name

        if current_url != 'django_glue_handler':
            clean_glue_session(request)
            request.session[settings.DJANGO_GLUE_SESSION_NAME] = dict()

        return None



from django.http import JsonResponse

JSON_RESPONSE_TYPES = (
    'success',
    'info',
    'warning',
    'error',
    'debug',
)


def generate_json_response(
        message_title,
        message_body,
        message_dict=None,
        response_status='200',
        response_type='success',
        additional_data=None
):
    if message_dict is None:
        message_dict = dict()

    if response_type not in JSON_RESPONSE_TYPES:
        raise ValueError(f'response_type "{response_type}" is not a valid, choices are {JSON_RESPONSE_TYPES}')

    return JsonResponse({
        'type': response_type,
        'message_title': message_title,
        'message_body': message_body,
        'message_dict': message_dict,
        'data': additional_data
    }, status=response_status)


def generate_json_404_response():
    return generate_json_response(
        message_title='Request not Found',
        message_body='The requested information, object or view you are looking for was not found.',
        response_status='404',
        response_type='error',
    )



from types import FunctionType
from time import time

from django.contrib.contenttypes.models import ContentType
from django.db.models import Model
from django.db.models.query import QuerySet

from django_glue.conf import settings
from django_glue.utils import encode_query_set_to_str, generate_field_dict

GLUE_ACCESS_TYPES = (
    'view',
    'add',
    'change',
    'delete',
)

GLUE_SESSION_TYPES = (
    'context',
    'query_set',
    'fields',
    'exclude',
)


def add_glue(
        request,
        unique_name: str,
        target, access: str,
        fields=('__all__',),
        exclude=('__none__',),
        **kwargs,
):
    if access in GLUE_ACCESS_TYPES:
        if isinstance(fields, (list, tuple)) and isinstance(exclude, (list, tuple)):
            if glue_unique_name_unused(request, unique_name):
                glue_session = get_glue_session(request)

                if isinstance(target, Model):
                    content_type = ContentType.objects.get_for_model(target)

                    glue_session['context'][unique_name] = {
                        'connection': 'model_object',
                        'access': access,
                        'app_label': content_type.app_label,
                        'model': content_type.model,
                        'object_id': target.pk,
                    }

                    glue_session['context'][unique_name]['fields'] = generate_field_dict(target, fields, exclude)

                elif isinstance(target, QuerySet):
                    content_type = ContentType.objects.get_for_model(target.query.model)

                    glue_session['context'][unique_name] = {
                        'connection': 'query_set',
                        'access': access,
                        'app_label': content_type.app_label,
                        'model': content_type.model,
                    }

                    glue_session['context'][unique_name]['fields'] = generate_field_dict(target.query.model(), fields,
                                                                                         exclude)
                    glue_session['query_set'][unique_name] = encode_query_set_to_str(target)

                elif isinstance(target, FunctionType):
                    pass

                else:
                    raise TypeError(
                        f'target is not a valid type must be a Python Method, Django Model or Django QuerySet')

                glue_session['fields'][unique_name] = fields
                glue_session['exclude'][unique_name] = exclude

                set_glue_keep_live(request, unique_name)
            else:
                raise ValueError(f'unique_name "{unique_name}" is already being used.')
        else:
            raise TypeError(f'fields or exclude is not a valid type must be a list or tuple')
    else:
        raise TypeError(f'access "{access}" is not a valid, choices are {GLUE_ACCESS_TYPES}')


def clean_glue_session(request):
    glue_keep_live_session = get_glue_keep_live_session(request)

    expired_session_list = list()

    expired_unique_name_set = set()
    active_unique_name_set = set()

    for key, val in glue_keep_live_session.items():
        if time() > val['expire_time']:
            expired_unique_name_set.update(val['unique_name_list'])
            expired_session_list.append(key)
        else:
            active_unique_name_set.update(val['unique_name_list'])

    for expired_session in expired_session_list:
        glue_keep_live_session.pop(expired_session)

    removable_unique_name_set = expired_unique_name_set.difference(active_unique_name_set)

    for unique_name in removable_unique_name_set:
        purge_unique_name_from_glue_session(request, unique_name)


def get_glue_session(request):
    request.session.setdefault(settings.DJANGO_GLUE_SESSION_NAME, dict())
    for session_type in GLUE_SESSION_TYPES:
        request.session[settings.DJANGO_GLUE_SESSION_NAME].setdefault(session_type, dict())

    return request.session[settings.DJANGO_GLUE_SESSION_NAME]


def get_glue_keep_live_next_expire_time():
    return time() + settings.DJANGO_GLUE_KEEP_LIVE_TIME


def get_glue_keep_live_session(request):
    return request.session.setdefault(settings.DJANGO_GLUE_KEEP_LIVE_SESSION_NAME, dict())


def glue_access_check(access, access_level):
    if GLUE_ACCESS_TYPES.index(access) >= GLUE_ACCESS_TYPES.index(access_level):
        return True
    else:
        return False


def glue_unique_name_unused(request, unique_name):
    if unique_name in get_glue_session(request)['context']:
        return False
    else:
        return True


def glue_run_method(request, model_object, method):
    if hasattr(model_object, method):
        getattr(model_object, method)(request)


def purge_unique_name_from_glue_session(request, unique_name):
    glue_session = get_glue_session(request)
    for session_type in GLUE_SESSION_TYPES:
        if unique_name in glue_session[session_type]:
            glue_session[session_type].pop(unique_name)


def set_glue_keep_live(request, unique_name):
    glue_keep_live_session = get_glue_keep_live_session(request)

    glue_keep_live_session.setdefault(
        request.path,
        {
            'expire_time': get_glue_keep_live_next_expire_time(),
            'unique_name_list': [],
        }
    )

    if unique_name not in glue_keep_live_session[request.path]['unique_name_list']:
        glue_keep_live_session[request.path]['unique_name_list'].append(unique_name)

def update_glue_keep_live(request, keep_live_path):
    glue_keep_live_session = get_glue_keep_live_session(request)
    if keep_live_path in glue_keep_live_session:
        glue_keep_live_session[keep_live_path]['expire_time'] = get_glue_keep_live_next_expire_time()




from django.contrib import admin
from django.urls import path, include
from django.views.generic import TemplateView

from tests import views

urlpatterns = [
    path("", views.TestView.as_view(), name="test"),
    path('admin/', admin.site.urls),
    path("glue/", include('django_glue.urls', namespace='django_glue')),
    path("no_glue/", TemplateView.as_view(template_name='no_glue.html'), name="no_glue"),
]


import logging

from django.views.generic import TemplateView

from tests.models import TestModel
from tests.utils import generate_randomized_test_model, generate_big_test_model
from django_glue.glue import add_glue


class TestView(TemplateView):
    template_name = 'test.html'

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)

        test_model_object = generate_randomized_test_model()

        logging.warning(f'Added TestModel object.')

        add_glue(self.request, 'test_model_1', test_model_object, 'change', exclude=('birth_date', 'anniversary_datetime',))
        add_glue(self.request, 'test_model_2', test_model_object, 'change', exclude=('birth_date', 'anniversary_datetime'))
        add_glue(self.request, 'test_model_3', test_model_object, 'delete', exclude=('birth_date', 'anniversary_datetime'))
        add_glue(self.request, 'test_model_4', test_model_object, 'change', exclude=('birth_date', 'anniversary_datetime'))
        logging.warning('Added model object glue for TestModel Object in write mode')

        add_glue(self.request, 'test_query_1', TestModel.objects.filter(id__gte=1).filter(id__lte=10000), 'change', exclude=('birth_date', 'anniversary_datetime'))
        add_glue(self.request, 'test_query_2', TestModel.objects.filter(id__gte=1).filter(id__lte=10000), 'add', exclude=('birth_date', 'anniversary_datetime'))
        add_glue(self.request, 'test_query_3', TestModel.objects.filter(id__gte=1).filter(id__lte=10000), 'delete', exclude=('birth_date', 'anniversary_datetime'))
        logging.warning('Added model query set glue for TestModel Object in read mode')

        big_test_model_object = generate_big_test_model()

        logging.warning(f'Added BigTestModel object.')

        return context_data



import logging, json
from abc import ABC, abstractmethod

from django.contrib.contenttypes.models import ContentType
from django.forms import model_to_dict
from django.http import Http404

from django_glue.json import generate_json_response, generate_json_404_response
from django_glue.glue import glue_run_method, glue_access_check, get_glue_session
from django_glue.utils import process_and_save_field_value, process_and_save_form_values, decode_query_set_from_str, \
    generate_simple_field_dict
from django_glue.conf import settings


class GlueRequestHandler(ABC):
    def __init__(self, request):
        self.request = request

        if request.content_type == 'application/json':
            logging.warning(request.body.decode('utf-8'))
            self.body_data = json.loads(request.body.decode('utf-8'))
        else:
            raise Http404

    @abstractmethod
    def process_response(self):
        pass

class GlueDataRequestHandler(GlueRequestHandler):
    def __init__(self, request):
        super().__init__(request)

        self.unique_name = self.body_data['unique_name']

        self.glue_session = get_glue_session(request)
        self.context = self.glue_session['context']

        self.method = request.method

        self.is_model_object = False
        self.is_query_set = False

        if self.unique_name in self.context:
            self.model_class = ContentType.objects.get_by_natural_key(
                self.context[self.unique_name]['app_label'],
                self.context[self.unique_name]['model']).model_class()
            self.access = self.context[self.unique_name]['access']

            if self.context[self.unique_name]['connection'] == 'model_object':
                self.is_model_object = True
                try:
                    self.model_object = self.model_class.objects.get(id=self.context[self.unique_name]['object_id'])
                except self.model_class.DoesNotExist:
                    self.model_object = None

            elif self.context[self.unique_name]['connection'] == 'query_set':
                self.is_query_set = True
                self.query_set = decode_query_set_from_str(self.glue_session['query_set'][self.unique_name])

        else:
            raise Http404

    def process_response(self):
        if self.is_model_object:
            if self.method == 'DELETE' and glue_access_check(self.access, 'delete'):
                return self.delete_model_object_handler()
            elif self.method == 'QUERY' and glue_access_check(self.access, 'view'):
                return self.query_model_object_handler()
            elif self.method == 'POST' and glue_access_check(self.access, 'add'):
                return self.post_model_object_handler()
            elif self.method == 'PUT' and glue_access_check(self.access, 'change'):
                return self.put_model_object_handler()
            else:
                return generate_json_404_response()
        elif self.is_query_set:
            if self.method == 'DELETE' and glue_access_check(self.access, 'delete'):
                return self.delete_query_set_handler()
            elif self.method == 'QUERY' and glue_access_check(self.access, 'view'):
                return self.query_query_set_handler()
            elif self.method == 'POST' and glue_access_check(self.access, 'add'):
                return self.post_query_set_handler()
            elif self.method == 'PUT' and glue_access_check(self.access, 'change'):
                return self.put_query_set_handler()
            else:
                return generate_json_404_response()
        else:
            return generate_json_404_response()

    def delete_model_object_handler(self):
        if self.model_object is not None:
            glue_run_method(self.request, self.model_object, 'django_glue_delete')

            if settings.DJANGO_GLUE_AUTO_DELETE:
                self.model_object.delete()

            return generate_json_response(
                'Delete Object Success',
                'The object you are trying to delete was completed successful'
            )
        else:
            return generate_json_404_response()

    def query_model_object_handler(self):
        if self.model_object is not None:
            glue_run_method(self.request, self.model_object, 'django_glue_view')

            return generate_json_response(
                'Retrieve Object Data Success',
                'The object data you are trying to receive was successful',
                additional_data=generate_simple_field_dict(
                    self.model_object,
                    self.glue_session['fields'][self.unique_name],
                    self.glue_session['exclude'][self.unique_name],
                )
            )
        else:
            return generate_json_404_response()

    def post_model_object_handler(self):
        new_model_object = self.model_class()
        process_and_save_form_values(new_model_object, self.body_data['data']['form_values'])

        glue_run_method(self.request, new_model_object, 'django_glue_create')

        return generate_json_response(
            'Create Object Success',
            'The thing you tried to create was completely successful',
            additional_data=generate_simple_field_dict(
                new_model_object,
                self.glue_session['fields'][self.unique_name],
                self.glue_session['exclude'][self.unique_name],
            )
        )

    def put_model_object_handler(self):
        if self.model_object is not None:

            if 'form_values' in self.body_data['data']:
                process_and_save_form_values(
                    self.model_object,
                    self.body_data['data']['form_values'],
                    self.glue_session['fields'][self.unique_name],
                    self.glue_session['exclude'][self.unique_name],
                )

            elif 'field_name' in self.body_data['data']:
                if self.body_data['data']['field_name'] in self.context[self.unique_name]['fields']:
                    process_and_save_field_value(
                        self.model_object,
                        self.body_data['data']['field_name'],
                        self.body_data['data']['value'],
                        self.glue_session['fields'][self.unique_name],
                        self.glue_session['exclude'][self.unique_name],
                    )

            glue_run_method(self.request, self.model_object, 'django_glue_update')

            return generate_json_response(
                'Update Object Successful',
                'The thing you tried to update was completely successful'
            )
        else:
            return generate_json_404_response()

    def delete_query_set_handler(self):
        if 'id' in self.body_data['data']:
            try:
                model_object = self.query_set.get(id=self.body_data['data']['id'])
            except self.model_class.DoesNotExist:
                logging.error('Handler: Object not Found')
                return generate_json_404_response()

            glue_run_method(self.request, model_object, 'django_glue_delete')

            if settings.DJANGO_GLUE_AUTO_DELETE:
                model_object.delete()

            return generate_json_response(
                'Delete Object in Query Set Success',
                'The thing you tried to delete was successful'
            )
        else:
            logging.error('Handler: ID not in Body')
            return generate_json_404_response()

    def query_query_set_handler(self):
        if 'id' in self.body_data['data']:
            id = int(self.body_data['data']['id'])
            if id > 0:
                try:
                    model_object = self.query_set.get(id=id)
                except self.model_class.DoesNotExist:
                    return generate_json_404_response()

                return generate_json_response(
                    'Retrieve Object Data in Query Set Success',
                    'The object data in query set you are trying to receive was successful',
                    additional_data=generate_simple_field_dict(
                        model_object,
                        self.glue_session['fields'][self.unique_name],
                        self.glue_session['exclude'][self.unique_name],
                    )
                )
            else:
                model_object_list = dict()
                for model_object in self.query_set:
                    model_object_list[model_object.id] = model_to_dict(model_object)

                return generate_json_response(
                    'View Query Set Success',
                    'The thing you tried to create was completely successful',
                    additional_data=model_object_list
                )
        else:
            return generate_json_404_response()

    def post_query_set_handler(self):
        new_model_object = self.model_class()
        process_and_save_form_values(
            new_model_object,
            self.body_data['data']['form_values'],
            self.glue_session['fields'][self.unique_name],
            self.glue_session['exclude'][self.unique_name],
        )

        glue_run_method(self.request, new_model_object, 'django_glue_create')

        return generate_json_response(
            'Create Object in Queryset Success',
            'The thing you tried to create was completely successful',
            additional_data=generate_simple_field_dict(
                new_model_object,
                self.glue_session['fields'][self.unique_name],
                self.glue_session['exclude'][self.unique_name],
            )
        )

    def put_query_set_handler(self):
        if 'id' in self.body_data['data']:
            try:
                model_object = self.query_set.get(id=self.body_data['data']['id'])
            except self.model_class.DoesNotExist:
                return generate_json_404_response()

            if 'form_values' in self.body_data['data']:
                process_and_save_form_values(
                    model_object,
                    self.body_data['data']['form_values'],
                    self.glue_session['fields'][self.unique_name],
                    self.glue_session['exclude'][self.unique_name],
                )

            elif 'field_name' in self.body_data['data']:
                if self.body_data['data']['field_name'] in self.context[self.unique_name]['fields']:
                    process_and_save_field_value(model_object, self.body_data['data']['field_name'],
                                                 self.body_data['data']['value'])

            glue_run_method(self.request, model_object, 'django_glue_update')

            return generate_json_response(
                'Update Successful',
                'The thing you tried to update was completely successful'
            )
        else:
            return generate_json_404_response()